/*
 * Copyright (C) 2012-2013, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"
/*
 * This is a semaphore-like API with explicit queueing and activation,
 * so contended scheduling/wakeups happen via epoll/kqueue and there
 * is never blocking of threads (other than the mutex)
 *
 * The main operations are mog_ioq_ready and mog_ioq_next
 *
 * mog_ioq_next is automatically called when releases a regular file.
 */
__thread struct mog_ioq *mog_ioq_current;

void mog_ioq_init(struct mog_ioq *ioq, struct mog_svc *svc, size_t val)
{
	ioq->cur = val;
	ioq->max = val;
	ioq->svc = svc;
	ioq->contended = false;
	SIMPLEQ_INIT(&ioq->ioq_head);
	CHECK(int, 0, pthread_mutex_init(&ioq->mtx, NULL));
}

/*
 * we do not need this, yet, but this will allow us to have multi-threaded
 * shutdown in the future (we currently drop into single-threaded mode)
 */
void mog_ioq_requeue_prepare(struct mog_ioq *ioq)
{
	assert(ioq->cur >= ioq->max &&
		"we should only get here when idle before mog_fdmap_requeue");

	SIMPLEQ_INIT(&ioq->ioq_head);
}

/*
 * this is only a hint, so no explicit memory barriers or atomics
 */
static inline void ioq_set_contended(struct mog_ioq *ioq)
{
	ioq->contended = true;
}

/*
 * This is like sem_trywait.  Each thread is only allowed to acquire
 * one ioq at once.
 *
 * client_mfd is the client socket, not the open (regular) file
 */
bool mog_ioq_ready(struct mog_ioq *ioq, struct mog_fd *client_mfd)
{
	bool good;

	assert(mog_ioq_current == NULL && "already holding mog_ioq_current");
	CHECK(int, 0, pthread_mutex_lock(&ioq->mtx));

	good = ioq->cur > 0;
	if (good) {
		/*
		 * assume the worst when we are the last one to
		 * acquire a free slot
		 */
		if (--ioq->cur == 0)
			ioq_set_contended(ioq);

		mog_ioq_current = ioq;
	} else {
		client_mfd->ioq_blocked = 1;
		SIMPLEQ_INSERT_TAIL(&ioq->ioq_head, client_mfd, ioqent);
		ioq_set_contended(ioq);
	}

	CHECK(int, 0, pthread_mutex_unlock(&ioq->mtx));

	return good;
}

/*
 * analogous to sem_post, this wakes up the next waiter
 * check_ioq may be NULL, if non-NULL, it validates against mog_ioq_current
 */
void mog_ioq_next(struct mog_ioq *check_ioq)
{
	struct mog_fd *client_mfd = NULL;

	if (mog_ioq_current == NULL)
		return;

	CHECK(int, 0, pthread_mutex_lock(&mog_ioq_current->mtx));

	assert((check_ioq == NULL) ||
	       (check_ioq == mog_ioq_current && "ioq mismatch (tls vs check)"));

	mog_ioq_current->cur++;
	if (mog_ioq_current->cur <= mog_ioq_current->max) {
		/* wake up any waiters */
		client_mfd = SIMPLEQ_FIRST(&mog_ioq_current->ioq_head);
		if (client_mfd) {
			SIMPLEQ_REMOVE_HEAD(&mog_ioq_current->ioq_head, ioqent);

			/* if there's another head, we're still contended */
			if (SIMPLEQ_FIRST(&mog_ioq_current->ioq_head))
				ioq_set_contended(mog_ioq_current);
		}
	} else {
		/* mog_ioq_adjust was called and lowered our capacity */
		mog_ioq_current->cur--;
		ioq_set_contended(mog_ioq_current);
	}
	CHECK(int, 0, pthread_mutex_unlock(&mog_ioq_current->mtx));

	/* wake up the next sleeper on this queue */
	if (client_mfd)
		mog_activeq_push(mog_ioq_current->svc->queue, client_mfd);

	mog_ioq_current = NULL;
}

/*
 * Returns true if the currently held ioq is contended.
 * This releases the contended flag if it is set, so the caller
 * is expected to yield the current thread shortly afterwards.
 * This is only a hint.
 */
bool mog_ioq_contended(void)
{
	struct mog_ioq *cur = mog_ioq_current;

	/* assume contended for non /devXXX* paths */
	if (!cur)
		return true;

	/*
	 * we only want to minimize the threads hitting true, so we use
	 * an atomic exchange and hope for the best.  This is only a hint.
	 */
	return __sync_bool_compare_and_swap(&cur->contended, true, false);
}

/*
 * called by the main/notify thread if the user has ever set
 * "server aio_threads = XX" via sidechannel.
 */
void mog_ioq_adjust(struct mog_ioq *ioq, unsigned value)
{
	assert(value > 0 && "mog_ioq_adjust value must be non-zero");
	CHECK(int, 0, pthread_mutex_lock(&ioq->mtx));
	ioq->max = value;

	/* capacity reduced, get some threads to yield themselves */
	if (ioq->cur > ioq->max)
		ioq_set_contended(ioq);

	CHECK(int, 0, pthread_mutex_unlock(&ioq->mtx));
}

void mog_ioq_destroy(struct mog_ioq *ioq)
{
	CHECK(int, 0, pthread_mutex_destroy(&ioq->mtx));
}

bool mog_ioq_unblock(struct mog_fd *mfd)
{
	if (mfd->ioq_blocked == 0)
		return false;

	mfd->ioq_blocked = 0;
	return true;
}
